@@ -35,6 +35,9 @@ module Network.Wai.Handler.WarpTLS (
     tlsSettingsChainRef,
     CertSettings,
 
+    -- ** Dynamically retrieved
+    tlsSettingsSni,
+
     -- * Accessors
     tlsCredentials,
     tlsLogging,
@@ -56,10 +59,22 @@ module Network.Wai.Handler.WarpTLS (
 ) where
 
 import Control.Applicative ((<|>))
+import Control.Exception (
+    Exception,
+    IOException,
+    SomeException (..),
+    bracket,
+    finally,
+    fromException,
+    handle,
+    handleJust,
+    onException,
+    throwIO,
+    try,
+ )
 import Control.Monad (guard, void)
 import qualified Data.ByteString as S
 import qualified Data.ByteString.Lazy as L
-import Data.Default.Class (def)
 import qualified Data.IORef as I
 import Data.Streaming.Network (bindPortTCP, safeRecv)
 import Data.Typeable (Typeable)
@@ -74,6 +89,7 @@ import Network.Socket (
 #endif
     withSocketsDo,
  )
+import qualified Control.Exception as E
 import Network.Socket.BufferPool
 import Network.Socket.ByteString (sendAll)
 import qualified Network.TLS as TLS
@@ -83,23 +99,7 @@ import Network.Wai.Handler.Warp
 import Network.Wai.Handler.Warp.Internal
 import Network.Wai.Handler.WarpTLS.Internal
 import System.IO.Error (ioeGetErrorType, isEOFError)
-import UnliftIO.Exception (
-    Exception,
-    IOException,
-    SomeException (..),
-    bracket,
-    finally,
-    fromException,
-    handle,
-    handleAny,
-    handleJust,
-    onException,
-    throwIO,
-    try,
- )
-import qualified UnliftIO.Exception as E
-import UnliftIO.Concurrent (newEmptyMVar, putMVar, takeMVar, forkIOWithUnmask)
-import UnliftIO.Timeout (timeout)
+import System.Timeout (timeout)
 
 ----------------------------------------------------------------
 
@@ -164,6 +164,22 @@ tlsSettingsChainMemory cert chainCerts key =
         { certSettings = CertFromMemory cert chainCerts key
         }
 
+-- | Smart constructor for TLS settings that obtains its credentials during
+-- Server Name Indication. Can be used to return different credentials
+-- depending on the hostname but also to retrieve dynamically updated
+-- credentials from an IORef. Credentials can be loaded from PEM-encoded chain
+-- and key files using 'TLS.credentialLoadX509'.
+--
+-- @since 3.4.13
+tlsSettingsSni :: (Maybe TLS.HostName -> IO TLS.Credentials) -> TLSSettings
+tlsSettingsSni onServerNameIndicationHook =
+  defaultTlsSettings
+    { tlsCredentials = Just (TLS.Credentials [])
+    , tlsServerHooks = (tlsServerHooks defaultTlsSettings)
+      { TLS.onServerNameIndication =  onServerNameIndicationHook
+      }
+    }
+
 -- | A smart constructor for 'TLSSettings', but uses references to in-memory
 -- representations of the certificate and key based on 'defaultTlsSettings'.
 --
@@ -179,6 +195,8 @@ tlsSettingsRef cert key =
         { certSettings = CertFromRef cert [] key
         }
 
+{-# DEPRECATED tlsSettingsRef "This function was added to allow Warp to serve new certificates without restarting, but it has always behaved the same as 'tlsSettingsMemory'. It will be removed in the next major release. To retain existing behavior, swich to 'tlsSettingsMemory'. To dynamically update credentials, see 'tlsSettingsSni'." #-}
+
 -- | A smart constructor for 'TLSSettings', but uses references to in-memory
 -- representations of the certificate and key based on 'defaultTlsSettings'.
 --
@@ -196,6 +214,8 @@ tlsSettingsChainRef cert chainCerts key =
         { certSettings = CertFromRef cert chainCerts key
         }
 
+{-# DEPRECATED tlsSettingsChainRef "This function was added to allow Warp to serve new certificates without restarting, but it has always behaved the same as 'tlsSettingsChainMemory'. It will be removed in the next major release. To retain existing behavior, swich to 'tlsSettingsChainMemory'. To dynamically update credentials, see 'tlsSettingsSni'." #-}
+
 ----------------------------------------------------------------
 
 -- | Running 'Application' with 'TLSSettings' and 'Settings'.
@@ -259,7 +279,7 @@ runTLSSocket' tlsset@TLSSettings{..} set credentials mgr sock =
   where
     get = getter tlsset set sock params
     params =
-        def -- TLS.ServerParams
+        TLS.defaultParamsServer
             { TLS.serverWantClientCert = tlsWantClientCert
             , TLS.serverCACertificates = []
             , TLS.serverDHEParams = tlsServerDHEParams
@@ -278,12 +298,12 @@ runTLSSocket' tlsset@TLSSettings{..} set credentials mgr sock =
                     <|> (if settingsHTTP2Enabled set then Just alpn else Nothing)
             }
     shared =
-        def
+        TLS.defaultShared
             { TLS.sharedCredentials = credentials
             , TLS.sharedSessionManager = mgr
             }
     supported =
-        def -- TLS.Supported
+        TLS.defaultSupported
             { TLS.supportedVersions = tlsAllowedVersions
             , TLS.supportedCiphers = tlsCiphers
             , TLS.supportedCompressions = [TLS.nullCompression]
@@ -324,12 +344,8 @@ mkConn
     -> params
     -> IO (Connection, Transport)
 mkConn tlsset set s params = do
-    var <- newEmptyMVar
-    _ <- forkIOWithUnmask $ \umask -> do
-        let tm = settingsTimeout set * 1000000
-        mct <- umask (timeout tm recvFirstBS)
-        putMVar var mct
-    mbs <- takeMVar var
+    let tm = settingsTimeout set * 1000000
+    mbs <- timeout tm recvFirstBS
     case mbs of
       Nothing -> throwIO IncompleteHeaders
       Just bs -> switch bs
@@ -342,6 +358,17 @@ mkConn tlsset set s params = do
 
 ----------------------------------------------------------------
 
+isAsyncException :: Exception e => e -> Bool
+isAsyncException e =
+    case E.fromException (E.toException e) of
+        Just (E.SomeAsyncException _) -> True
+        Nothing -> False
+
+throughAsync :: IO a -> SomeException -> IO a
+throughAsync action (SomeException e)
+  | isAsyncException e = E.throwIO e
+  | otherwise          = action
+
 httpOverTls
     :: TLS.TLSParams params
     => TLSSettings
@@ -350,7 +377,7 @@ httpOverTls
     -> S.ByteString
     -> params
     -> IO (Connection, Transport)
-httpOverTls TLSSettings{..} _set s bs0 params =
+httpOverTls TLSSettings{..} set s bs0 params =
     makeConn `onException` close s
   where
     makeConn = do
@@ -359,16 +386,21 @@ httpOverTls TLSSettings{..} _set s bs0 params =
         let recvN = wrappedRecvN rawRecvN
         ctx <- TLS.contextNew (backend recvN) params
         TLS.contextHookSetLogging ctx tlsLogging
-        TLS.handshake ctx
-        mysa <- getSocketName s
-        attachConn mysa ctx
-    wrappedRecvN recvN n = handleAny (const mempty) $ recvN n
+        let tm = settingsTimeout set * 1000000
+        mconn <- timeout tm $ do
+            TLS.handshake ctx
+            mysa <- getSocketName s
+            attachConn mysa ctx
+        case mconn of
+          Nothing -> throwIO IncompleteHeaders
+          Just conn -> return conn
+    wrappedRecvN recvN n = handle (throughAsync (return "")) $ recvN n
     backend recvN =
         TLS.Backend
             { TLS.backendFlush = return ()
 #if MIN_VERSION_network(3,1,1)
             , TLS.backendClose =
-                gracefulClose s 5000 `E.catch` \(SomeException _) -> return ()
+                gracefulClose s 5000 `E.catch` throughAsync (return ())
 #else
             , TLS.backendClose = close s
 #endif
